# Fishnet Simulator

# FishnetSim implements Fishnet in a single address space,
# by simulating message passing between the various nodes in the net.
#
# We are created by main; in turn we create k simulated "Nodes",
# running student code.  The student code running on each node sends packets, 
# receives packets, and gets keyboard
# input, just as it would if they were running on real hardware.
#
# We use a global event queue to keep track of which message to deliver
# next; messages are delivered in the same order they would be
# delivered in real time.
#
# One bit of unrealism -- code executes infinitely quickly in simulated time.
# In practice, for very fast networks, processing time can be a bottleneck
# that would need to be explicitly considered.
#
# Another bit of unrealism -- we only squash events to be delivered to a node
# if the node (or the incoming edge in the case of a packet) is failed at the 
# time when the event is generated or when it would be delivered.  
# This is (at least sometimes) incorrect in the case where the node is alive when the
# event is queued, then the node (or edge) fails,
# and then is restarted before the event would have be delivered.
# In this case we deliver it; equally reasonable would be to squash it.

require 'manager'

# Windows doesn't support select() on stdin, so make a note if we're on Windows.
require 'rbconfig.rb'
$WINDOWS = (Config::CONFIG['host'] =~ /mswin32/)


module Fishnet

# Fishnet implemented as a simulator
class Simulator < Manager
	@@MaxToSimulate = MaxAddress - 1  # without the broadcast address
	
	# Initialize the fishnet simulation
	# 	x -- number of Nodes to simulate
	# 	topoFile -- topology file to use for connecting the nodes
	def initialize(x, topoFile)
		raise ArgumentError if x <= 0 or x > @@MaxToSimulate
	
		@now = 0				# simulated time, in microseconds
		@events = SortedEvents.new	# events occuring in the future

		@timescale = 1  # fish time passes at 1*real time.

		@nodes = {}
		0.upto(x) { |i|
			#proxy = SimulatorProxy.new(self, i)
			@nodes[i] = Node.new(self, i)
		}
		
		# parse the topology file, and defer processing if we see a "time" command		
		@cmds = SimulationCmds.new	# command line parser
		@defer = @cmds.parseFile(topoFile, 0)
		@events.push(@defer, proc { parseRestOfFile }) if @defer
		super()
	end
	
	attr_reader :nodes
	attr_accessor :timescale
	
	# process the remainder of the topology commands file
	def parseRestOfFile()
		@defer = @cmds.parseRemainder(@now)
		@events.push(@defer, proc { parseRestOfFile }) if @defer
	end
	
	# Start each of the nodes; we're ready for nodes to start sending and
	# receiving packets.  Once we've started everybody, start processing the
	# event queue, delivering each event in turn.  First check the keyboard to
	# see if there are commands to deliver.
	def start
		log Debug::PHYSICAL, "Fishnet started. Topology: #{$topology}." +
		      (@timescale == 1 ? "" : " Timescale: #{@timescale}.")
		@nodes.each_value {|x| x.start}
		
		if $WINDOWS
			inputs = []
		else
			inputs = [$stdin]
		end

		@now = 1		# we don't process keyboard messages until time > 0
		inputDefer = nil

		loop do
			if $fishInput
				if $fishInput.eof?
					$fishInput = nil
				elsif (!inputDefer or inputDefer <= @now)
					l = $fishInput.gets
					inputDefer = @cmds.processLine(l, @now) || @now

					# register a dummy event so our event loop wakes up to read the next
					# line.
					@events.push(inputDefer, proc {})
				end
			end

			nextevent = @events.first
			if nextevent
				deltatime = nextevent.timeToOccur - @now if nextevent
				if deltatime > 0
					waitTime = deltatime*@timescale/1_000_000.0
				else
					waitTime = 0
				end
			else
				# no event to wait for, so wait indefinitely for input.
				waitTime = nil
			end

			# wait for either input or the next event to fire.
			beforeSelectTime = fishTime()
			selectret = select(inputs, nil, nil, waitTime)
			if selectret   # if there is input, we process it.
				if @timescale > 0
					@now += ((fishTime() - beforeSelectTime)/@timescale).to_i
				end
				inputs = selectret[0]
				inputs.each { |input|
					line = input.gets
					@cmds.processLine(line, @now) if line
				}
				next # and then go back to waiting for an event.
			end

			# if we get here, we have already waited the appropriate amount of
			# real time.
			# so it's safe to fast-forward @now.
			@now = nextevent.timeToOccur if nextevent.timeToOccur > @now

			# and finally, run all pending events that should happen.
			while nextevent and nextevent.timeToOccur <= @now
				ev = @events.shift
				ev.upcall.call
				nextevent = @events.first
			end
		end
	end
	
	# called by Node to send a pkt to a peer.  Need to check the topology
	# to make sure we have a connection to the peer, and to expand the message
	# list if this is a broadcast address.
	def sendPkt(from, to, pkt)
		super(to, pkt)		# check arguments
		
		if to == BroadcastAddress then
			@nodes.each { |addr, node|
				edge = $topology.liveEdge?(from, addr)
				#puts "edge(#{addr}, #{node}) -> edge"
				deliverPkt(addr, node, from, pkt, edge) if edge
			}
			@nodes[from].onReceive(from, pkt)
		elsif edge = $topology.liveEdge?(from, to)
			deliverPkt(to, @nodes[to], from, pkt, edge)
		else	# should we return an error here, or fail silently?
			log Debug::PHYSICAL, "Cannot send packet from #{from} to #{to}: no connection."
		end
	end
	
	# schedule a packet to be delivered to the appropriate node
	def deliverPkt(to, tonode, from, pkt, edge)
		toOccur = edge.schedulePkt(from, pkt.length, @now)
		return unless toOccur   # don't schedule this if the packet got dropped.
		@events.push(toOccur, proc {
			# don't deliver message if link or endpoint has failed
			if $topology.liveEdge?(from, to) then
				tonode.onReceive(from, pkt)
			end
		})
	end
	
	# called by node to register a wakeup alarm at time t
	# @now is in microseconds
	def addTimerAt(addr, t, callback)
		super(addr, t, callback)		# check arguments
		@events.push(t, proc {
			# the node may have failed in the meantime
			callback.call(@now/1000) if $topology.liveNode?(addr)
		})
	end

	# The node has failed, and needs to be restarted
	# Just allocate a new node, and kick it.  
	def restartNode(x)
		#proxy = SimulatorProxy.new(self, x)
		@nodes[x] = Node.new(self, x)
		@nodes[x].start()
	end
	
	# retrieve current time in milliseconds
	def now
		@now / 1000
	end

	def node_log(debug_level, node, *args)
		super debug_level, node, "#{node}: ", *args
	end

end  # FishnetSim

end # module Fishnet

# vim: set ts=2 sw=2 noet :
