# Node -- file defining the data structures and code for the
# protocol stack for a node participating in the fishnet

# This code is under student control; we provide a baseline
# "hello world" example just to show how it interfaces with the FishNet classes.
#
# NOTE: per-node global variables are prohibited.
#  In simulation mode, we execute many nodes in parallel, and thus
#  they would share any globals.  Please put per-node state in class Node
#  instead.
#
# In other words, all the code defined here should be generic -- it shouldn't
#   matter to this code whether it is executing in a simulation,
#   in an emulation, or on the iPAQs.

# Print all debuginng messages.
$debugFlags = Debug::ALL

# This class is used to maintain information about Ping requests made by 
# this node.
class PingReq
  def initialize(to, msg, now)
    @to = to
    @msg = msg
    @sent = now
  end
  
  attr_reader :to, :msg, :sent
end

# Class Node -- the network protocol software for a single CPU.
#   Note that this code must be written as a "state machine" --
#  each upcall must do its work and return, so that other upcalls
#  can be delivered.
class Node < Fishnet::NodeBase
  WAKEUP = 10000      # wakeup every 10 seconds

  def initialize(fishnet, addr)
    super(fishnet, addr)
    @pinged = []          # the nodes that I've ping'ed
  end
  
  # called when fishnet is ready for this node to run.
  # All we do is start a timer to look for un-acknowledged ping's.
  def start()
    addTimer(WAKEUP, proc { onTimer })
  end

  # call this to send a packet
  def send(to, pkt)
    log Debug::PHYSICAL, "Sending to #{to}: #{pkt.inspect}"
    sendPkt(to, pkt.pack)
  end
  
  # called when a packet has arrived for this node
  def onReceive(from, str)
    pkt = Fishnet::Packet.unpack(str)
    log Debug::PHYSICAL, "Receiving: ", pkt, "\n"

    if !pkt
      log Debug::PHYSICAL, "Illegal packet #{str.inspect}"
    else
      receivePacket(pkt)
    end
  end

  # processes packets destined for this node
  def receivePacket(pkt)
    case pkt.protocol
    when Fishnet::PingPkt
      receivePing(pkt)
    when Fishnet::PingReplyPkt
      receivePingReply(pkt)
    else
      log Debug::PHYSICAL, "Unknown protocol #{pkt.protocol}"
    end
  end
  
  # if packet is a ping, reply with the contents of the packet  
  def receivePing(pkt)
    $fishnet.pong  # play the sound card when someone pings us!
    log Debug::APPLICATION, "Received ping from #{pkt.src}: #{pkt.msg.inspect}"

    reply = Fishnet::Packet.new(pkt.src, @addr, Fishnet::PingReplyPkt, 
                                Fishnet::MaxTTL, 0, pkt.msg)
    send(pkt.src, reply)
  end

  # if packet is a ping reply, check that it matches what was sent    
  def receivePingReply(pkt)
    found = nil
    @pinged.each do |x|
      found = x if (pkt.src == x.to) and
        (pkt.msg == x.msg)
      end
    if found
      log Debug::APPLICATION, "Ping reply from #{pkt.src}: #{pkt.msg.inspect}"
      @pinged.delete found
    else
      log Debug::APPLICATION, "Unexpected ping reply from #{pkt.src}: #{pkt.msg.inspect}"
    end
  end 

  # called when the user types a command to this node
  # for now, send keyboard message as a ping to a neighbor  
  def onCommand(command)
    if command =~ /\A\s*(\d+):\s*(.*)/ then
      to, msg = $1.to_i, $2

      pkt = Fishnet::Packet.new(to, @addr, Fishnet::PingPkt, 
                                Fishnet::MaxTTL, 0, msg) 
      send(to, pkt)
      @pinged.push PingReq.new(to, msg, @fishnet.now)

    else
      log Debug::INPUT, "onCommand: Unrecognized command: #{command}"
    end
  end
  
  # called when Node requested a wakeup, and the time
  # is now >= the alarm time (note that the wakeup can
  # sometimes occur a bit later than requested)
  # for now, look to see if there are any unanswered pings.
  # Note that this doesn't necessarily signal an error, since the ping
  # or ping reply could have been dropped.
  def onTimer
    @pinged.delete_if do |x|
      if x.sent + 10000 < @fishnet.now then
	log Debug::APPLICATION, "Message undelivered to #{x.to} after 10 seconds\n"
        true
      else
        false
      end
    end  
    addTimer(WAKEUP, proc { onTimer })
  end

end
